import os
import webbrowser
import dill
import tempfile
import vipy.math
from vipy.util import remkdir, tempdir
import builtins
import logging as python_logging
import warnings


# Global mutable dictionary
GLOBAL = {'VERBOSE': True,       # If False, will silence everything, equivalent to calling vipy.globals.silent()
          'VERBOSITY': 2,        # 0=debug, 1=warn, 2=info, only if VERBOSE=True
          'DASK_CLIENT': None,   # Global Dask() client for distributed processing
          'CACHE':os.environ['VIPY_CACHE'] if 'VIPY_CACHE' in os.environ else None,   # Cache directory for vipy.video and vipy.image donwloads
          'GPU':None,            # GPU index assigned to this process
          'LOGGING':False,       # If True, use python logging (handler provided by end-user) intead of print 
          'LOGGER':None,         # The global logger used by vipy.globals.print() and vipy.globals.warn() if LOGGING=True
          'GUI':{'escape':False},
          'AWS':{'AWS_ACCESS_KEY_ID':os.environ['VIPY_AWS_SECRET_ACCESS_KEY'] if 'VIPY_AWS_ACCESS_KEY_ID' in os.environ else None,
                 'AWS_SECRET_ACCESS_KEY':os.environ['VIPY_AWS_SECRET_ACCESS_KEY'] if 'VIPY_AWS_SECRET_ACCESS_KEY' in os.environ else None,
                 'AWS_SESSION_TOKEN':os.environ['VIPY_AWS_SESSION_TOKEN'] if 'VIPY_AWS_SESSION_TOKEN' in os.environ else None},
          'LATEX':os.environ['VIPY_LATEX'] if 'VIPY_LATEX' in os.environ else None}



def logging(enable=None, format=None):
    """Single entry point for enabling/disabling logging vs. printing
       
       All vipy functions overload "from vipy.globals import print" for simplified readability of code.
       This global function redirects print or warn to using the standard logging module.
       If format is provided, this will create a basicConfig handler, but this should be configured by the end-user.    
    """
    if enable is not None:
        assert isinstance(enable, bool)
        GLOBAL['LOGGING'] = enable
        if format is not None:
            python_logging.basicConfig(level=python_logging.INFO, format=format)
        GLOBAL['LOGGER'] = python_logging.getLogger('vipy')
        GLOBAL['LOGGER'].propagate = True if enable else False
        
    return GLOBAL['LOGGING']


def warn(s):
    if GLOBAL['VERBOSE']:
        warnings.warn(s) if (not GLOBAL['LOGGING'] or GLOBAL['LOGGER'] is None) else GLOBAL['LOGGER'].warn(s)

        
def print(s, end='\n'):
    """Main entry point for all print statements in the vipy package. All vipy code calls this to print helpful messages.
      
       -Printing can be disabled by calling vipy.globals.silent()
       -Printing can be redirected to logging by calling vipy.globals.logging(True)
       -All print() statements in vipy.* are overloaded to call vipy.globals.print() so that it can be redirected to logging

    """
    if GLOBAL['VERBOSE']:
        builtins.print(s, end=end) if (not GLOBAL['LOGGING'] or GLOBAL['LOGGER'] is None) else GLOBAL['LOGGER'].info(s)


def verbose():
    """The global verbosity level, only really used right now for FFMPEG messages"""
    GLOBAL['VERBOSE'] = True

def isverbose():
    return GLOBAL['VERBOSE']

def silent():
    GLOBAL['VERBOSE'] = False    

def issilent():
    return GLOBAL['VERBOSE'] == False 

def verbosity(v):
    assert v in [0,1,2]    # debug, warn, info
    GLOBAL['VERBOSITY'] = v

def debug():
    verbose()
    verbosity(0)

def isdebug():
    return GLOBAL['VERBOSE'] and GLOBAL['VERBOSITY'] == 0


def cache(cachedir=None):
    """The cache is the location that URLs are downloaded to on your system.  This can be set here, or with the environment variable VIPY_CACHE"""
    if cachedir is not None:
        os.environ['VIPY_CACHE'] = remkdir(cachedir)
        GLOBAL['CACHE'] = cachedir
    return os.environ['VIPY_CACHE'] if 'VIPY_CACHE' in os.environ else None
    

def user_hit_escape(b=None):
    if b is None:
        if GLOBAL['GUI']['escape']:
            GLOBAL['GUI']['escape'] = False  # toggle it
            return True
        else:
            return False
    else:
        # Set in vipy.gui.using_matplotlib.escape_to_exit()
        assert isinstance(b, bool)
        GLOBAL['GUI']['escape'] = b  
            
def cpuonly():
    GLOBAL['GPU'] = None


def gpuindex(gpu=None):
    if gpu == 'cpu':
        cpuonly()
    elif gpu is not None:
        GLOBAL['GPU'] = gpu
    return GLOBAL['GPU']


def dask(num_processes=None, num_gpus=None, dashboard=False, address=None, pct=None):
    """Return the current Dask client, can be accessed globally for parallel processing.
    
       -pct: [0,1] the percentage of the current machine to use
       -address:  the dask scheduler of the form 'HOSTNAME:PORT'
       -num_processes:  the number of prpcesses to use on the current machine
       -num_gpus:  the number of GPUs to use on the current machine
    """
    from vipy.batch import Dask
    if pct is not None:
        assert pct > 0 and pct <= 1
        import multiprocessing
        num_processes = vipy.math.poweroftwo(pct*multiprocessing.cpu_count())        
    if (address is not None or (num_processes is not None and (GLOBAL['DASK_CLIENT'] is None or GLOBAL['DASK_CLIENT'].num_processes() != num_processes)) or num_gpus is not None):
        GLOBAL['DASK_CLIENT'] = Dask(num_processes, dashboard=dashboard, verbose=isverbose(), address=address, num_gpus=num_gpus)        
    return GLOBAL['DASK_CLIENT']

def parallel(n=None):
    """Enable parallel processing with n>=1 processes"""
    if n is None:
        return GLOBAL['DASK_CLIENT'].num_processes() if GLOBAL['DASK_CLIENT'] is not None else 0
    else:
        assert isinstance(n, int) and n>=1
        return dask(num_processes=n)
        
def noparallel():
    """Disable all parallel processing"""
    GLOBAL['DASK_CLIENT'] = None 

def nodask():
    """Alias for noparallel()"""
    return noparallel()



import os
from os import path
import urllib.parse as urlparse
from vipy.util import try_import
from vipy.globals import print

class Dropbox():

    _access_token = os.environ.get('VIPY_DROPBOX_ACCESS_TOKEN')
    _app_key = os.environ.get('VIPY_DROPBOX_APP_KEY')
    _app_secret = os.environ.get('VIPY_DROPBOX_APP_SECRET_KEY')

    def __init__(self):
        raise ValueError('FIXME: this uses an older version of dropbox')

        try_import('dropbox')
        import dropbox  # optional

        if self._access_token is None:
            self.link()

    def link(self):
        flow = dropboxapi.client.DropboxOAuth2FlowNoRedirect(self._app_key, self._app_secret)

        authorize_url = flow.start()
        print('1. Go to: ' + authorize_url)
        print('2. Click "Allow" (you might have to log in first)')
        print('3. Copy the authorization code.')
        code = raw_input("Enter the authorization code here: ").strip()

        # This will fail if the user enters an invalid authorization code
        access_token, user_id = flow.finish(code)
        self._access_token = access_token
        return access_token

    def put(self, filename, folder=None, verbose=True):
        client = dropbox.client.DropboxClient(self._access_token)
        if verbose:
            print('linked account: ', client.account_info())

        if folder is None:
            dropbox_path = '/' + path.basename(filename)
        else:
            dropbox_path = '/' + str(folder) + '/' + path.basename(filename)

        f = open(filename)
        response = client.put_file(dropbox_path, f)
        if verbose:
            print('uploaded: ', response)

        folder_metadata = client.metadata('/')
        if verbose:
            print('metadata: ', folder_metadata)

        share = client.share(dropbox_path, short_url=False)

        # https://www.dropbox.com/help/201/en
        p = urlparse.urlparse(share['url'])
        public_url = urlparse.urlunsplit(('http','dl.dropboxusercontent.com',p[2],None,None))
        # public_url = public_url + '?dl=1'
        return public_url

    def get(self, filename):
        client = dropbox.client.DropboxClient(self._access_token)
        f, metadata = client.get_file_and_metadata('/' + filename)
        cachefile = filename
        out = open(cachefile, 'w')
        out.write(f.read())
        out.close()
        print(metadata)
        return cachefile



import os
from glob import glob
from vipy.util import remkdir, try_import, dirlist, imlist, filetail
from vipy.image import ImageDetection, ImageCategory
try_import('scipy.io', 'scipy')
from scipy.io import loadmat


URL = 'http://www.cslab.openu.ac.il/download/wolftau/YouTubeFaces.tar.gz'


class YouTubeFaces(object):
    def __init__(self, datadir):
        self.datadir = remkdir(datadir)
        self.ytfdir = datadir
        if not os.path.isdir(os.path.join(self.datadir, 'frame_images_DB')):
            raise ValueError('Download YouTubeFaces dataset with "wget %s -O %s; cd %s; tar zxvf YouTubeFaces.tar.gz", and initialize with YouTubeFace(datadir="%s/YouTubeFaces")' % (URL, os.path.join(self.datadir, 'YouTubeFaces.tar.gz'), self.datadir, self.datadir))

    def __repr__(self):
        return str('<viset.youtubefaces: %s>' % self.ytfdir)

    def subjects(self):
        return os.listdir(os.path.join(self.ytfdir, 'descriptors_DB'))

    def videos(self, subject):
        videos = {}
        for d in dirlist(os.path.join(self.ytfdir, 'frame_images_DB', subject)):
            k_videoindex = filetail(d)
            videos[k_videoindex] = []
            for f in imlist(d):
                videos[k_videoindex].append(ImageCategory(filename=f, category=subject))
            videos[k_videoindex] = sorted(videos[k_videoindex], key=lambda im: im.filename())
        return videos

    def parse(self, subject):
        """Parse youtubefaces into a list of ImageDetections"""

        # Write images and annotations
        # The data in this file is in the following format:
        # filename,[ignore],x,y,width,height,[ignore],[ignore]
        # where:
        # x,y are the center of the face and the width and height are of the rectangle that the face is in.
        # For example:
        # $ head -3 Richard_Gere.labeled_faces.txt
        # Richard_Gere\3\3.618.jpg,0,262,165,132,132,0.0,1
        # Richard_Gere\3\3.619.jpg,0,260,164,131,131,0.0,1
        # Richard_Gere\3\3.620.jpg,0,261,165,129,129,0.0,1

        imlist = []
        categorydir = os.path.join(self.ytfdir, 'frame_images_DB')
        for i,infilename in enumerate(glob('%s/*labeled_faces.txt' % categorydir)):
            print('[vipy.dataset.youtubefaces:] parsing "%s" ' % infilename)
            with open(infilename, 'r') as infile:
                for line in infile:
                    (imname, ignore, x_ctr, y_ctr, w, h, ignore, ignore) = line.split(',')
                    imname = imname.replace('\\', '/')
                    category = imname.split('/')[0]
                    xmin = int(x_ctr) - int(float(w) / 2.0)
                    ymin = int(y_ctr) - int(float(h) / 2.0)
                    xmax = int(xmin) + int(w)
                    ymax = int(ymin) + int(h)
                    p = os.path.join(categorydir, imname)

                    (category,videoid,filename) = imname.split('/')
                    mediaID = '%s_%s' % (category, videoid)
                    imlist.append(ImageDetection(filename=p, category=category, xmin=int(xmin), ymin=int(ymin), xmax=int(xmax), ymax=int(ymax)).setattribute('MEDIA_ID', str(mediaID)))

        # Done!
        self._parsed = imlist
        return imlist

    def splits(self):
        mat = loadmat(os.path.join(self.ytfdir, 'meta_data', 'meta_and_splits.mat'))
        return mat['Splits']  # splits are indexed into mat['mat_names'] dictionary, parse me!



import os
import numpy as np
from vipy.util import dirlist, imlist, readcsv, filebase, readlist, groupbyasdict
from vipy.image import ImageDetection


class VGGFace2(object):
    def __init__(self, datadir, seed=None):
        assert os.path.isdir(os.path.join(datadir, 'n000001')) and os.path.exists(os.path.join(datadir, 'identity_meta.csv')), 'Download and unpack VGGFace2 data and metadata (http://www.robots.ox.ac.uk/~vgg/data/vgg_face2/) to "%s"' % datadir

        self.datadir = datadir
        self._subjects = None
        if seed is not None:
            np.random.seed(seed)  # for repeatable np.random

    def __repr__(self):
        return str('<vipy.dataset.vggface2: %s>' % self.datadir)

    def subjects(self):
        if self._subjects is None:
            self._subjects = [filebase(d) for d in dirlist(self.datadir)]
        return self._subjects  # cached

    def wordnetid_to_name(self):
        csv = readcsv(os.path.join(self.datadir, 'identity_meta.csv'), ignoreheader=True)
        return {str(x[0]):str(x[1]).replace('"', '') for x in csv}

    def vggface2_to_vggface1(self):
        assert os.path.exists(os.path.join(self.datadir, 'class_overlap_vgg1_2.txt')), 'Download class_overlap_vgg1_2.txt to "%s"' % self.datadir
        csv = readcsv(os.path.join(self.datadir, 'class_overlap_vgg1_2.txt'), separator=' ', ignoreheader=True)
        return {x[0]:x[1] for x in csv}

    def name_to_wordnetid(self):
        d = self.wordnetid_to_name()
        return {v:k for (k,v) in d.items()}

    def names(self):
        return list(self.wordnetid_to_name().values())

    def trainset(self):
        assert os.path.exists(os.path.join(self.datadir, 'train_list.txt')), 'Download "train_list.txt" from http://www.robots.ox.ac.uk/~vgg/data/vgg_face2/ to "%s"' % self.datadir
        csv = readlist(os.path.join(self.datadir, 'train_list.txt'))
        return [os.path.join(self.datadir, x).strip() for x in csv]

    def testset(self):
        assert os.path.exists(os.path.join(self.datadir, 'test_list.txt')), 'Download "test_list.txt" from http://www.robots.ox.ac.uk/~vgg/data/vgg_face2/ to "%s"' % self.datadir
        csv = readlist(os.path.join(self.datadir, 'test_list.txt'))
        return [os.path.join(self.datadir, x).strip() for x in csv]

    def split(self, f):
        """Convert absolute path /path/to/subjectid/filename.jpg from training or testing set to (subjectid, filename.jpg)"""
        x = os.path.split(f)
        subjectid = os.path.split(x[-2])[-1]
        imagefile = x[-1]
        return (subjectid, imagefile)

    def frontalset(self, n_frontal=1):
        # http://www.robots.ox.ac.uk/~vgg/data/vgg_face2/meta/test_posetemp_imglist.txt
        assert(n_frontal >= 1 and n_frontal <= 10)
        assert os.path.exists(os.path.join(self.datadir, 'test_posetemp_imglist.txt')), 'Download "test_posetemp_imglist.txt" from (http://www.robots.ox.ac.uk/~vgg/data/vgg_face2/meta/test_posetemp_imglist.txt) to "%s"' % self.datadir
        d = groupbyasdict([x.strip().split('/') for x in readlist(os.path.join(self.datadir, 'test_posetemp_imglist.txt'))], lambda v: v[0])
        d_subjectid_to_frontallist = {k:[os.path.join(self.datadir, k, y[1]) for y in v[0:n_frontal]] for (k,v) in d.items()}  # first and second set of five are frontal
        for (k,v) in d_subjectid_to_frontallist.items():
            for f in v:
                yield ImageDetection(filename=f).category(k)

    def dataset(self):
        """Return a generator to iterate over dataset"""
        for d in dirlist(os.path.join(self.datadir)):
            for f in imlist(d):
                yield ImageDetection(filename=f).category(filebase(d))

    def fastset(self):
        """Return a generator to iterate over dataset"""
        for d in dirlist(os.path.join(self.datadir)):
            for f in imlist(d):
                yield ImageDetection(filename=f, category=filebase(d))

    def take(self, n, wordnetid=None):
        """Randomly select n images from the dataset, or n images of a given subjectid"""
        subjectid = np.random.choice(self.subjects(), n) if wordnetid is None else [wordnetid] * n
        takelist = []
        for s in subjectid:
            d = os.path.join(self.datadir, s)
            f = np.random.choice(imlist(d),1)[0]
            im = ImageDetection(filename=f).category(filebase(d))
            takelist.append(im)
        return takelist

    def take_per_subject(self, n):
        """Randomly select n images per subject from the dataset"""
        subjectid = self.subjects()
        takelist = []
        for s in subjectid:
            d = os.path.join(self.datadir, s)
            for k in range(0,n):
                f = np.random.choice(imlist(d),1)[0]
                im = ImageDetection(filename=f).category(filebase(d))
                takelist.append(im)
        return takelist

    def subjectset(self, wordnetid):
        """Iterator for single subject"""
        assert wordnetid in self.wordnetid_to_name().keys(), 'Invalid wordnetid "%s"' % wordnetid
        d = os.path.join(self.datadir, wordnetid)
        for f in imlist(d):
            yield ImageDetection(filename=f, category=filebase(d))



MAJOR = 1
MINOR = 9
RELEASE = 8
VERSION = '%d.%d.%d' % (MAJOR, MINOR, RELEASE)


def num(versionstring=VERSION):
    (major, minor, release) = versionstring.split('.')    
    return 100*100*int(major) + 100*int(minor) + int(release)


def at_least_version(versionstring):
    """Is versionstring='X.Y.Z' at least the current version?"""
    return num(VERSION) >= num(versionstring)


def is_at_least(versionstring):
    """Synonym for at_least_version"""
    return num(VERSION) >= num(versionstring)    


def at_least_major_version(major):
    return MAJOR >= int(major)


